@@ -1,6 +1,6 @@
 #!/usr/bin/env perl
 #
-#    Copyright (C) 2012-2020 Genome Research Ltd.
+#    Copyright (C) 2012-2022 Genome Research Ltd.
 #
 #    Author: Petr Danecek <pd3@sanger.ac.uk>
 #
@@ -32,6 +32,7 @@ use File::Temp qw/ tempfile tempdir /;
 use IO::Handle;
 
 my $opts = parse_params();
+srand($$opts{seed});
 
 test_bgzip($opts, 0);
 test_bgzip($opts, 4);
@@ -59,6 +60,7 @@ test_rebgzip($opts);
 test_logging($opts);
 test_plugin_loading($opts);
 test_realn($opts);
+test_bcf_set_variant_type($opts);
 
 print "\nNumber of tests:\n";
 printf "    total   .. %d\n", $$opts{nok}+$$opts{nfailed};
@@ -79,6 +81,7 @@ sub error
         "Usage: test.pl [OPTIONS]\n",
         "Options:\n",
         "   -r, --redo-outputs              Recreate expected output files.\n",
+        "   -s, --random-seed <int>         Initialise rand() with a different seed.\n",
         "   -t, --temp-dir <path>           When given, temporary files will not be removed.\n",
         "   -f, --fail-fast                 Fail-fast mode: exit as soon as a test fails.\n",
         "   -h, -?, --help                  This help message.\n",
@@ -104,12 +107,13 @@ sub safe_tempdir
 
 sub parse_params
 {
-    my $opts = { keep_files=>0, nok=>0, nfailed=>0 };
+    my $opts = { keep_files=>0, nok=>0, nfailed=>0, seed=>42 };
     my $help;
     Getopt::Long::Configure('bundling');
     my $ret = GetOptions (
             't|temp-dir:s' => \$$opts{keep_files},
             'r|redo-outputs' => \$$opts{redo_outputs},
+            's|random-seed=i' => \$$opts{seed},
             'f|fail-fast' => \$$opts{fail_fast},
             'h|?|help' => \$help
             );
@@ -520,6 +524,32 @@ sub test_view
     my ($opts, $nthreads) = @_;
     my $tv_args = $nthreads ? "-\@$nthreads" : "";
 
+    # Files appropriate for CRAM V3.1 and V4.0 testing
+    my %cram31 = ("auxf#values.sam"   => 1,
+                  "c1#pad3.sam"       => 1,
+                  "ce#5.sam"          => 1,
+                  "ce#1000.sam",      => 1,
+                  "ce#large_seq.sam", => 1,
+                  "ce#supp.sam",      => 1,
+                  "xx#MD.sam",        => 1,
+                  "xx#blank.sam",     => 1,
+                  "xx#large_aux.sam", => 1,
+                  "xx#pair.sam",      => 1,
+                  "xx#tlen.sam"       => 1);
+
+    # Files appropriate for CRAM multi-ref containers
+    my %cram_ms = ("ce#1000.sam"      => 1,
+                   "ce#5.sam"         => 1,
+                   "ce#5b.sam"        => 1,
+                   "ce#unmap.sam"     => 1,
+                   "ce#unmap1.sam"    => 1,
+                   "ce#unmap2.sam"    => 1,
+                   "xx#blank.sam"     => 1,
+                   "xx#minimal.sam"   => 1,
+                   "xx#tlen.sam"      => 1,
+                   "xx#tlen2.sam"     => 1,
+                   "xx#triplet.sam"   => 1);
+
     foreach my $sam (glob("*#*.sam")) {
         my ($base, $ref) = ($sam =~ /((.*)#.*)\.sam/);
         $ref .= ".fa";
@@ -536,9 +566,11 @@ sub test_view
         $test_view_failures = 0;
 
         # SAM -> BAM -> SAM
-        testv $opts, "./test_view $tv_args -S -b $sam > $bam";
-        testv $opts, "./test_view $tv_args $bam > $bam.sam_";
-        testv $opts, "./compare_sam.pl $sam $bam.sam_";
+        if ($sam eq "ce#1000.sam") {
+            testv $opts, "./test_view $tv_args -S -b $sam > $bam";
+            testv $opts, "./test_view $tv_args $bam > $bam.sam_";
+            testv $opts, "./compare_sam.pl $sam $bam.sam_";
+        }
 
         # SAM -> BAMu -> SAM
         testv $opts, "./test_view $tv_args -S -l0 -b $sam > $bam";
@@ -550,94 +582,53 @@ sub test_view
         testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
         testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
 
-        # BAM -> CRAM2 -> BAM -> SAM
-        $cram = "$bam.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=2.1 $bam > $cram";
-        testv $opts, "./test_view $tv_args -b -D $cram > $cram.bam";
-        testv $opts, "./test_view $tv_args $cram.bam > $cram.bam.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.bam.sam_";
-
         # SAM -> CRAM3u -> SAM
-        $cram = "$base.tmp.cram";
-        testv $opts, "./test_view $tv_args -t $ref -S -l0 -C -o VERSION=3.0 $sam > $cram";
-        testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
+        if ($sam eq "ce#1000.sam") {
+            $cram = "$base.tmp.cram";
+            testv $opts, "./test_view $tv_args -t $ref -S -l0 -C -o VERSION=3.0 $sam > $cram";
+            testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
+            testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
+        }
 
-        # BAM -> CRAM3 -> BAM -> SAM
+        # BAM -> CRAM3 -> SAM
         $cram = "$bam.cram";
         testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 $bam > $cram";
-        testv $opts, "./test_view $tv_args -b -D $cram > $cram.bam";
-        testv $opts, "./test_view $tv_args $cram.bam > $cram.bam.sam_";
+        testv $opts, "./test_view $tv_args $cram > $cram.bam.sam_";
         testv $opts, "./compare_sam.pl $md $sam $cram.bam.sam_";
 
-        # CRAM3 -> CRAM2
-        $cram = "$base.tmp.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=2.1 $cram > $cram.cram";
-
-        # CRAM2 -> CRAM3
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 $cram.cram > $cram";
-
         # CRAM3 -> CRAM3 + multi-slice
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 -o seqs_per_slice=7 -o slices_per_container=5 $cram.cram > $cram";
-        testv $opts, "./test_view $tv_args $cram > $cram.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
-
-        ## Experimental CRAM 3.1 support.
-        # SAM -> CRAM31u -> SAM
-        foreach my $profile (qw/fast normal small archive/) {
-            $cram = "$base.tmp.cram";
-            testv $opts, "./test_view $tv_args -t $ref -S -l7 -C -o VERSION=3.1 -o $profile $sam > $cram";
-            testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
+        if (exists($cram_ms{$sam}) && $nthreads > 0) {
+            testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 -o seqs_per_slice=7 -o slices_per_container=5 $cram > $cram.ms";
+            testv $opts, "./test_view $tv_args $cram.ms > $cram.sam_";
             testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
         }
 
-        # BAM -> CRAM31 -> BAM -> SAM
-        $cram = "$bam.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.1 $bam > $cram";
-        testv $opts, "./test_view $tv_args -b -D $cram > $cram.bam";
-        testv $opts, "./test_view $tv_args $cram.bam > $cram.bam.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.bam.sam_";
+        if (exists($cram31{$sam}) && $nthreads > 0) {
+            ## Experimental CRAM 3.1 support.
+            # SAM -> CRAM31 -> SAM
+            my @p = $sam eq "ce#1000.sam"
+                ? (qw/fast normal small archive/)
+                : (qw/archive/);
+            foreach my $profile (@p) {
+                $cram = "$base.tmp.cram";
+                testv $opts, "./test_view $tv_args -t $ref -S -l7 -C -o VERSION=3.1 -o $profile $sam > $cram";
+                testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
+                testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
+            }
 
-        # CRAM31 -> CRAM30
-        $cram = "$base.tmp.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 $cram > $cram.cram";
-
-        # CRAM30 -> CRAM31
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.1 $cram.cram > $cram";
-
-        # CRAM31 -> CRAM31 + multi-slice
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.1 -o seqs_per_slice=7 -o slices_per_container=5 $cram.cram > $cram";
-        testv $opts, "./test_view $tv_args $cram > $cram.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
-
-        ## Experimental CRAM 4.0 support.
-        # SAM -> CRAM40u -> SAM
-        foreach my $profile (qw/fast normal small archive/) {
-            $cram = "$base.tmp.cram";
-            testv $opts, "./test_view $tv_args -t $ref -S -l7 -C -o VERSION=4.0 -o $profile $sam > $cram";
-            testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
-            testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
+            ## Experimental CRAM 4.0 support.
+            # SAM -> CRAM40 -> SAM
+            @p = $sam eq "ce#large_seq.sam" || $sam eq "xx#large_aux.sam"
+                ? (qw/fast normal small archive/)
+                : (qw/archive/);
+            foreach my $profile (@p) {
+                $cram = "$base.tmp.cram";
+                testv $opts, "./test_view $tv_args -t $ref -S -l7 -C -o VERSION=4.0 -o $profile $sam > $cram";
+                testv $opts, "./test_view $tv_args -D $cram > $cram.sam_";
+                testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
+            }
         }
 
-        # BAM -> CRAM40 -> BAM -> SAM
-        $cram = "$bam.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=4.0 $bam > $cram";
-        testv $opts, "./test_view $tv_args -b -D $cram > $cram.bam";
-        testv $opts, "./test_view $tv_args $cram.bam > $cram.bam.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.bam.sam_";
-
-        # CRAM40 -> CRAM30
-        $cram = "$base.tmp.cram";
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=3.0 $cram > $cram.cram";
-
-        # CRAM30 -> CRAM40
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=4.0 $cram.cram > $cram";
-
-        # CRAM40 -> CRAM40 + multi-slice
-        testv $opts, "./test_view $tv_args -t $ref -C -o VERSION=4.0 -o seqs_per_slice=7 -o slices_per_container=5 $cram.cram > $cram";
-        testv $opts, "./test_view $tv_args $cram > $cram.sam_";
-        testv $opts, "./compare_sam.pl $md $sam $cram.sam_";
-
         # Java pre-made CRAM -> SAM
         my $jcram = "${base}_java.cram";
         if (-e $jcram) {
@@ -646,6 +637,14 @@ sub test_view
             testv $opts, "./compare_sam.pl -Baux $md $sam $jsam";
         }
 
+        # embed_ref=2 mode
+        my $ersam = "ce#1000.sam";
+        my $ercram = "ce#1000_er.tmp.cram";
+        my $ersam2 = "${ercram}.sam";
+        testv $opts, "./test_view $tv_args -C -p $ercram $ersam";
+        testv $opts, "./test_view $tv_args -p $ersam2 $ercram";
+        testv $opts, "./compare_sam.pl $ersam $ersam2";
+
         if ($test_view_failures == 0)
         {
             passed($opts, "$sam conversions");
@@ -853,6 +852,31 @@ sub test_index
         $wtmp =~ s/\//\\\\/g;
     }
     test_cmd($opts,out=>'tabix.out',cmd=>"$$opts{bin}/tabix $wtmp/index.vcf.gz##idx##$wtmp/index.vcf.gz.tbi 1:10000060-10000060");
+
+    cmd("$$opts{path}/test_view -b -p $$opts{tmp}/index2.bam -x $$opts{tmp}/index2.bam.bai $$opts{path}/index2.sam");
+    for (my $tid = 1; $tid <= 2; $tid++) {
+        for (my $pos = 1; $pos <= 2; $pos++) {
+            # All queries should return exactly two sequences.
+            # The input data consists of mapped/unmapped and unmapped/mapped
+            # in both orders.
+            # Done verbatim as test_cmd cannot return $out for us to check.
+            my $test = "$$opts{path}/test_view $$opts{tmp}/index2.bam $tid:${pos}000000-${pos}000000";
+            print "test_index:\n\t$test\n";
+            my ($ret, $out) = _cmd($test);
+            if ($ret ne 0) {
+                failed($opts, $test);
+            } else {
+                my $rnum = ($out =~ s/^[^@].*\n//gm);
+                if ($rnum ne 2) {
+                    failed($opts, $test);
+                } else {
+                    passed($opts, $test);
+                }
+            }
+        }
+    }
+    unlink("$$opts{tmp}/index2.bam");
+    unlink("$$opts{tmp}/index2.bam.bai");
 }
 
 sub test_bcf2vcf
@@ -962,7 +986,7 @@ sub test_bcf_sr_sort
     my ($opts, %args) = @_;
     for (my $i=0; $i<10; $i++)
     {
-        my $seed = int(rand(time));
+        my $seed = int(rand(100000000));
         my $test = 'test-bcf-sr';
         my $cmd  = "$$opts{path}/test-bcf-sr.pl -t $$opts{tmp} -s $seed";
         print "$test:\n";
@@ -1040,3 +1064,17 @@ sub test_realn {
     # Revert quality values (using data in ZQ tags)
     test_cmd($opts, cmd => "$test_realn -f $$opts{path}/realn02.fa -i $$opts{path}/realn02_exp-a.sam -o -", out => "realn02_exp.sam");
 }
+
+sub test_bcf_set_variant_type
+{
+    my ($opts) = @_;
+    my $test = 'test-bcf_set_variant_type';
+    my $cmd  = "$$opts{path}/test-bcf_set_variant_type";
+    print "$test:\n";
+    print "\t$cmd\n";
+    my ($ret,$out) = _cmd($cmd);
+    if ( $ret ) {
+        print $out;
+        failed($opts,$test);
+    } else { passed($opts,$test); }
+}
